@@ -18,6 +18,7 @@
 # Contact: Gene Shuman <gene@valimail.com>
 #
 
+import importlib.resources
 import re
 import sys
 import copy
@@ -28,7 +29,6 @@ from authres.arc import ARCAuthenticationResult
 from authres.dmarc import DMARCAuthenticationResult
 from dkim import ARC, DKIM, arc_verify, dkim_verify, DKIMException, rfc822_parse
 from dns.exception import DNSException
-from pkg_resources import resource_filename  # Part of setuptools
 
 # Please accept my appologies for doing this
 try:
@@ -57,17 +57,18 @@ def check_psddmarc_list(psdname, dnsfunc=dns_query):
     """Check psddmarc.org list of PSD DMARC participants"""
     try:
         # If the PSD registry is locally available, use it.
-        psdfile_name = resource_filename('authheaders', 'psddmarc.csv')
-        psd_file = open(psdfile_name)
-        psds = []
-        for line in psd_file.readlines():
-            sp = line.split(',')
-            if sp[1] == 'current':
-                psds.append(sp[0][1:])
-        if psdname in psds:
-            return True
-        else:
-            return False
+        ref = importlib.resources.files('authheaders') / 'psddmarc.csv'
+        with importlib.resources.as_file(ref) as psdfile_name:
+            with open(psdfile_name) as psd_file:
+                psds = []
+                for line in psd_file.readlines():
+                    sp = line.split(',')
+                    if sp[1] == 'current':
+                        psds.append(sp[0][1:])
+                if psdname in psds:
+                    return True
+                else:
+                    return False
     except:
         # If not, use the DNS query list.
         psd_list_host = '.psddmarc.org'
@@ -231,7 +232,10 @@ def dmarc_per_from(from_domain, spf_result=None, dkim_result=None, dnsfunc=None,
         if not policy_only and spf_result and spf_result.result == "pass":
             # The domain in SPF results often includes the local part, even though
             # generally it SHOULD NOT (RFC 7601, Section 2.7.2, last paragraph).
-            mail_from_domain = get_domain_part(spf_result.smtp_mailfrom)
+            try:
+                mail_from_domain = get_domain_part(spf_result.smtp_mailfrom)
+            except IndexError:
+                mail_from_domain = None
             spf_result.smtp_mailfrom = mail_from_domain
             if aspf == "s" and from_domain == mail_from_domain:
                 result = "pass"
@@ -312,7 +316,7 @@ def check_arc(msg, logger=None, dnsfunc=None):
     return ARCAuthenticationResult(result=cv.decode('ascii'), result_comment=comment)
 
 
-def check_dmarc(msg, spf_result=None, dkim_result=None, dnsfunc=None, psddmarc=False):
+def check_dmarc(msg, spf_result=None, dkim_result=None, dnsfunc=None, psddmarc=False, dmarcbis=False):
 
     # get from domain
     headers, _ = rfc822_parse(msg)
@@ -322,8 +326,17 @@ def check_dmarc(msg, spf_result=None, dkim_result=None, dnsfunc=None, psddmarc=F
         # multi-from processing per RFC 7489 6.6.1
         domain_results = []
         for from_header in from_headers:
-            from_domain = get_domain_part(from_header)
-            domain_results.append(dmarc_per_from(from_domain, spf_result, dkim_result, dnsfunc, psddmarc))
+            try:
+                from_domain = get_domain_part(from_header)
+            except IndexError:
+                result = 'permerror'
+                result_comment = 'Unable to extract From domain: {0}'.format(from_header)
+                return DMARCAuthenticationResult(result=result, result_comment=result_comment, header_from='none', policy='none')
+            try:
+                domain_results.append(dmarc_per_from(from_domain, spf_result=spf_result, dkim_result=dkim_result, dnsfunc=dnsfunc, psddmarc=psddmarc, dmarcbis=dmarcbis))
+            except dmarc_lookup.DMARCException as result_comment:
+                result = 'permerror'
+                return DMARCAuthenticationResult(result=result, result_comment=result_comment, header_from=from_domain, policy='none')
 
         for domain in domain_results:
             if domain[3] == 'reject':
@@ -340,8 +353,18 @@ def check_dmarc(msg, spf_result=None, dkim_result=None, dnsfunc=None, psddmarc=F
         result, result_comment, from_domain, policy = domain
     elif len(from_headers) == 1:
         from_header =  from_headers[0]
-        from_domain = get_domain_part(from_header)
-        result, result_comment, from_domain, policy = dmarc_per_from(from_domain, spf_result, dkim_result, dnsfunc, psddmarc)
+        try:
+            from_domain = get_domain_part(from_header)
+        except IndexError:
+            result = 'permerror'
+            result_comment = 'Unable to extract From domain: {0}'.format(from_header)
+            return DMARCAuthenticationResult(result=result, result_comment=result_comment, header_from='none', policy='none')
+        try:
+            result, result_comment, from_domain, policy = dmarc_per_from(from_domain, spf_result=spf_result, dkim_result=dkim_result, dnsfunc=dnsfunc, psddmarc=psddmarc, dmarcbis=dmarcbis)
+        except dmarc_lookup.DMARCException as result_comment:
+            result = 'permerror'
+            return DMARCAuthenticationResult(result=result, result_comment=result_comment, header_from=from_domain, policy='none')
+
     else:
         result = 'none'
 
@@ -351,7 +374,7 @@ def check_dmarc(msg, spf_result=None, dkim_result=None, dnsfunc=None, psddmarc=F
         return DMARCAuthenticationResult(result=result, header_from=from_domain)
 
 
-def authenticate_message(msg, authserv_id, prev=None, spf=False, dkim=True, arc=False, dmarc=True, ip=None, mail_from=None, helo=None, dnsfunc=None, psddmarc=False):
+def authenticate_message(msg, authserv_id, prev=None, spf=False, dkim=True, arc=False, dmarc=True, ip=None, mail_from=None, helo=None, dnsfunc=None, psddmarc=False, dmarcbis=False):
     """Authenticate an RFC822 message and return the Authentication-Results header
     @param msg: an RFC822 formatted message (with either \\n or \\r\\n line endings)
     @param authserv_id: The id of the server performing the authentication
@@ -359,7 +382,8 @@ def authenticate_message(msg, authserv_id, prev=None, spf=False, dkim=True, arc=
     @param spf: Perform SPF check
     @param dkim: Perform DKIM check
     @param dmarc: Perform DMARC check
-    @param psddmarc: Perform PSD DMARC check (draft-ietf-dmarc-psd)
+    @param psddmarc: Perform PSD DMARC check (RFC 9091)
+    @param dmarcbis: Use DMARCbis policy discovery and alignment
     @param arc: Perform ARC chain validation check
     @param ip: (SPF) IP address of incoming request
     @param mail_from: (SPF) Sender declared in MAIL FROM
@@ -393,7 +417,7 @@ def authenticate_message(msg, authserv_id, prev=None, spf=False, dkim=True, arc=
         results.append(arc_result)
 
     if dmarc:
-        dmarc_result = check_dmarc(msg, spf_result, dkim_result, dnsfunc=dnsfunc, psddmarc=psddmarc)
+        dmarc_result = check_dmarc(msg, spf_result, dkim_result, dnsfunc=dnsfunc, psddmarc=psddmarc, dmarcbis=dmarcbis)
         results.append(dmarc_result)
 
     auth_res = AuthenticationResultsHeader(authserv_id=authserv_id, results=results)
